from scipy.stats import vonmises
import numpy as np
from ckmeans_1d_dp import ckmeans

class CircluarKMeans1dOptimal:


    def __init__(self, n_clusters):
        self.n_clusters= n_clusters

    def fit(self, X):
        labels_iter = []
        cluster_points_iter = []
        cluster_centers_iter = []
        distances_cluster_iter = []
        moved_data_iter = []
        errors = []

        for x in X:
            data = self.move_data(X, x)
            moved_data_iter.append(data)
            labels_ = ckmeans(data, k=self.n_clusters).cluster
            cluster_points = []
            cluster_points_moved = []
            for i in range(self.n_clusters):
                data = np.array(data)
                cluster_points_moved.append(data[labels_==i])
                cluster_points.append(X[labels_==i])
            cluster_centers_ = []
            errors_cl =[]
            
            dists_coll = []
            for i in range(len(cluster_points_moved)):
                cl =cluster_points[i]
                cl_moved =cluster_points_moved[i]
                std, mean_vm, n = vonmises.fit(cl)
                #mean = mean%(2*np.pi)
                #z_mean = np.sin(mean) + 1j*np.cos(mean)
                #cl_z = np.sin(cl) + 1j*np.cos(cl)
                #distances_cluster = np.abs([np.angle(z- z_mean) for z in cl_z])
                
                mean = cl_moved.mean()
                distances_cluster = np.abs(cl_moved-mean)**2

                cluster_centers_.append(mean_vm)
                dists_coll.append(distances_cluster)
                error = distances_cluster.mean()
                errors_cl.append(error)

            distances_cluster_iter.append(dists_coll) 
            errors_cl = np.array(errors_cl)
            sum_err = errors_cl.sum()
            errors.append(sum_err)     
            labels_iter.append(labels_)
            cluster_points_iter.append(cluster_points)
            cluster_centers_iter.append(cluster_centers_)

        index_min = np.argmin(errors, axis=-1)
        self.moved_data = moved_data_iter[index_min]
        self.moved_data_iter = moved_data_iter
        self.cluster_points = cluster_points_iter[index_min]
        self.cluster_points_iter = cluster_points_iter
        self.cluster_centers_ = cluster_centers_iter[index_min]
        self.cluster_centers_iter = cluster_centers_iter
        self.centroids = cluster_centers_iter[index_min]
        self.labels = labels_iter[index_min]
        self.labels_iter = labels_iter
        self.errors = errors
        self.distances_cluster_iter = distances_cluster_iter
        
    def move_data(self, X, x):
        data = ((X  - x))% (2 * np.pi)
        return data
    
    def _organize_in_cluster_points(self, X, clusters_indexes, k):
        cluster_points = []
        for i in range(0,k):
            indexes = self._all_indices(i,clusters_indexes)
            cl = np.array([X[j] for j in indexes])
            cluster_points.append(cl)
        return cluster_points
    
    def _all_indices(self, value, qlist):
        qlist = list(qlist)
        indices = []
        idx = -1
        while True:
            try:
                idx = qlist.index(value, idx+1)
                indices.append(idx)
            except ValueError:
                break
        return indices


import numpy as np
import scipy

from circular_clustering.kmeans_1d import KMeans1d
from circular_clustering.circular_interval import CircularInterval
from circular_clustering.optimal_circular_kmeans import CircluarKMeans1dOptimal
from circular_clustering.circular_k_means_weighted import CircularWeightedKMeans



class CircularXMeansQuantiles:


    def __init__(self, X, kmax = 20, confidence=0.99, use_optimal_k_means=True):
        self.X = X
        self.num = np.size(self.X, axis=0)
        self.dim = 1
        self.KMax = kmax
        self.confidence = confidence

        if use_optimal_k_means == True:
            self.circular_k_means = CircluarKMeans1dOptimal
        else:
            self.circular_k_means = CircularWeightedKMeans

    def probability_same_cluster(self,mu, sigma, a, b):
        
        return np.abs(scipy.stats.vonmises(loc=mu, kappa=sigma).cdf(b) - scipy.stats.vonmises(loc=mu, kappa=sigma).cdf(a))

    def get_quantiles(self, confidence, kappa, loc):

        a = scipy.stats.vonmises.ppf(1-confidence, kappa=kappa, loc=loc)
        b = scipy.stats.vonmises.ppf(confidence, kappa=kappa, loc=loc)
        return a, b

    def fit(self):
        k = self.KMax
        X = self.X

        while(1):
            ok = k

            #Improve Params
            kmeans = self.circular_k_means(n_clusters=k)
            kmeans.fit(X)
            X=X
            
            labels = kmeans.labels
            addk=0
            
            stop=False
            i=0

            while stop==False and i<=k:
                cli = X[labels == i]
                for j in range(k):
                    if j != i and j>i:    
                        clj = X[labels == j]
                
                        kappai, loci, n = scipy.stats.vonmises.fit(cli)
                        kappaj, locj, n = scipy.stats.vonmises.fit(clj)
            

                        ai, bi = self.get_quantiles(self.confidence, kappai, loci)
                        aj, bj = self.get_quantiles(self.confidence, kappaj, locj)

                        circular_interval_i = CircularInterval(ai, bi, gamma=loci)
                        circular_interval_j = CircularInterval(aj, bj, gamma=locj)
                        

                        overlap = circular_interval_i.intervals_intersect(circular_interval_j)
                        
                        if overlap:
                            addk -= 1
                            stop=True
                            break
                i=i+1
            k += addk


            if ok == k or k <1:
                break


        #Calculate labels and centroids
        kmeans = self.circular_k_means(n_clusters=k)
        kmeans.fit(X)
        self.labels = kmeans.labels
        self.k = k
        self.m = kmeans.cluster_centers_
        self.centroids = kmeans.centroids
        self.cluster_points = kmeans.cluster_points



 


import numpy as np
import random
from scipy.stats import vonmises

def circ_dist(point, data_pts, r):
        min_dists = []
        for data in data_pts:
            min_dist = np.min([np.abs(point - data), r - np.abs(point - data)])
            min_dists.append(min_dist)
        return min_dists

class CircularWeightedKMeans:
    def __init__(self, n_clusters, max_iter=300):
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.iteration = 0
        self.centroids = None
        self.prev_centroids = None
        self.std = None
        self.r = None
        self.fac = None
        self.clust_weights = None
        self.cluster_points = []
        self.labels = []
        
    def fit(self, X_train, weights=None, r=2*np.pi):
        self.r = r
        self.fac = 2*np.pi/self.r
        self.centroids = [random.choice(X_train)]
        weights = np.ones(len(X_train))
        for _ in range(self.n_clusters-1):
            dists = circ_dist(self.centroids[-1], X_train, self.r)
            dists /= np.sum(dists)
            new_centroid_idx, = np.random.choice(range(len(X_train)), size=1, p=dists)
            self.centroids += [X_train[new_centroid_idx]]
        while np.not_equal(self.centroids, self.prev_centroids).any() and self.iteration < self.max_iter:
            self.prev_centroids = self.centroids
            self.clust_weights = [0]*self.n_clusters
            self.std = [0]*self.n_clusters
            self.cluster_points = [[] for _ in range(self.n_clusters)]
            sorted_weights = [[] for _ in range(self.n_clusters)]
            for x, w in zip(X_train, weights):
                # assign each point to closest centroid
                dists = circ_dist(x, self.centroids, self.r)
                closest_centroid_idx = np.argmin(dists)
                self.clust_weights[closest_centroid_idx] += w
                self.cluster_points[closest_centroid_idx].append(x)
                sorted_weights[closest_centroid_idx].append(w)
            for idx, _ in enumerate(self.centroids):
                # weighted mean of all points in cluster gives new centroid
                self.std[idx], self.centroids[idx], n = vonmises.fit(np.array(self.cluster_points[idx]))

                self.clust_weights[idx] = np.sum(np.array(sorted_weights[idx]))
            self.clust_weights = [w/np.sum(self.clust_weights) for w in self.clust_weights]
            self.iteration += 1
            self.cluster_centers_ = len(self.centroids)
            self.get_labels(X_train)

    def get_labels(self,X_train): 
        label_list = list(range(len(self.cluster_points)))
        self.labels = np.zeros(len(X_train))
        for i in range(len(self.cluster_points)):
            cl = self.cluster_points[i] 
            label = label_list[i]
            for x in cl:
                index = np.where(X_train==x)[0][0]
                self.labels[index] = label
                
    
    def plot(self, X_train, weights):
        # Histogram 
        plt.hist(X_train, bins=20, weights=weights)
        sizes = self.clust_weights/np.max(self.clust_weights)*200
        # draw std as lines 
        for i, centroid in enumerate(self.centroids):
            # random color
            color = np.random.rand(3,)
            plt.scatter(centroid, max(weights)/2*sizes[i]/max(sizes), s=sizes[i], zorder=2, color=color, label=f'Cluster {i}, x={int(centroid)}, weight {self.clust_weights[i]/np.sum(self.clust_weights):.2f}, std {self.std[i]:.2f}')
            y_range = 2*[max(weights)/2*sizes[i]/max(sizes)]
            plt.plot([centroid-self.std[i], centroid+self.std[i]], y_range, zorder=1, color=color)
        plt.legend()
        plt.show()


import numpy as np
import matplotlib


class CircularInterval:
    a = None
    b = None
    sign = 1.


    def __init__(self, a, b, gamma=None, sign=1.):
        self.a = a
        self.b = b
        if gamma:
            self.sign = self.get_sign(a,gamma)
        else:
            self.sign = sign

    def get_sign(self, a,b):
        za = [np.cos(a), np.sin(a)]
        zb = [np.cos(b), np.sin(b)] 

        return np.sign(np.cross(za,zb))
    
    def intervals_intersect(self, circularInterval):
        a = circularInterval.a
        b = circularInterval.b
        return self.point_in_circular_interval(a) or self.point_in_circular_interval(b)

    
    def point_in_circular_interval(self, alpha):
        sign_a = self.get_sign(self.a, alpha)
        sign_b = self.get_sign(self.b, alpha)
        if self.sign == sign_a and self.sign != sign_b:
            return True
        else:
            return False
        
    def draw(self,ax):
        wedge = matplotlib.patches.Wedge((0,0), 1, 360*(1/(2*np.pi))*self.a,  360*(1/(2*np.pi))*self.b, alpha=0.5)
        ax.add_artist(wedge)
    
   






import numpy as np
from ckmeans_1d_dp import ckmeans

class KMeans1d:


    def __init__(self, n_clusters):
        self.n_clusters= n_clusters

    def fit(self, X):
        self.labels_ = ckmeans(X, k=self.n_clusters).cluster
        self.cluster_points = self._organize_in_cluster_points(X, self.labels_,self.n_clusters)
        self.cluster_centers_ = []
        for cl in self.cluster_points:
            self.cluster_centers_.append(cl.mean())
        
    
    def _organize_in_cluster_points(self, X, clusters_indexes, k):
        cluster_points = []
        for i in range(0,k):
            indexes = self._all_indices(i,clusters_indexes)
            cl = np.array([X[j] for j in indexes])
            cluster_points.append(cl)
        return cluster_points
    
    def _all_indices(self, value, qlist):
        qlist = list(qlist)
        indices = []
        idx = -1
        while True:
            try:
                idx = qlist.index(value, idx+1)
                indices.append(idx)
            except ValueError:
                break
        return indices